DIR_ACT=`pwd`

NAME=$1;

echo "=== $NAME ===";

DIR_DATA=${DIR_ACT}/${NAME}_BRILL;

BASE_NAME=${DIR_DATA}/${NAME};
TRAIN=${BASE_NAME}.train;
TRAIN_1=${TRAIN}.01;
TRAIN_2=${TRAIN}.02;
TRAIN_UNTAGGED_1=${BASE_NAME}.untagged.01;
TRAIN_UNTAGGED_2=${BASE_NAME}.untagged.02;
LIST_BIG=${BASE_NAME}.big;
LIST_SMALL=${BASE_NAME}.small;
BIGRAMS_TRAIN=${BASE_NAME}.bigrams;
BIGRAMS_ALL=${BIGRAMS_TRAIN}.all;
RULES_TAGS=${BASE_NAME}.rules.tags;
RULES_CONTEXTUAL=${BASE_NAME}.rules.contextual;
LEXICON_TRAIN=${BASE_NAME}.lexicon.train;
LEXICON_ALL=${BASE_NAME}.lexicon.all;
TAGGED_DUMMY=${BASE_NAME}.tagged.dummy;
TAGGED_RESULT=${BASE_NAME}.tagged.result;
TEST=${BASE_NAME}.test;
TEST_UNTAGGED=${TEST}.untagged;
WORDS=${BASE_NAME}.words;


RESULT=${BASE_NAME}.result

DIR_UTIL=${DIR_ACT}/RULE_BASED_TAGGER_V1.14/Utilities/
DIR_LEARNER=${DIR_ACT}/RULE_BASED_TAGGER_V1.14/Learner_Code/
DIR_BIN=${DIR_ACT}/RULE_BASED_TAGGER_V1.14/Bin_and_Data/

PATH=${PATH}:${DIR_BIN};

echo "BRILL - Copy files";
mkdir -p $DIR_DATA;
cp ${DIR_ACT}/${NAME}.train ${DIR_DATA}
cp ${DIR_ACT}/${NAME}.heldout ${DIR_DATA}
cp ${DIR_ACT}/${NAME}.test ${DIR_DATA}



echo "BRILL - Split training corpora";
cat ${TRAIN} | ${DIR_UTIL}/divide-in-two-rand.prl ${TRAIN_1} ${TRAIN_2}

echo "BRILL - Create untagged";
cat ${TRAIN_1} | $DIR_UTIL/tagged-to-untagged.prl > ${TRAIN_UNTAGGED_1};

echo "BRILL - Create big word list";
cat ${TRAIN_UNTAGGED_1} | ${DIR_UTIL}/wordlist-make.prl | sort -nr -k2 | cut -f1 -d" " > ${LIST_BIG};

echo "BRILL - Create tagged list";
cat ${TRAIN_1} | ${DIR_UTIL}/word-tag-count.prl | sort -nr -k3 > ${LIST_SMALL};

echo "BRILL - Create bigrams";
cat ${TRAIN_UNTAGGED_1} | ${DIR_UTIL}/bigram-generate.prl | awk '{ print $1,$2}' > ${BIGRAMS_TRAIN}

echo "BRILL - Learn rules";
if [ ! -f ${RULES_TAGS} ]; then
	${DIR_LEARNER}/unknown-lexical-learn.prl ${LIST_BIG} ${LIST_SMALL} ${BIGRAMS_TRAIN} 300 ${RULES_TAGS};
fi;

echo "BRILL - Prepare corpora for contextua rules";
cat ${TRAIN_1} | ${DIR_UTIL}/make-restricted-lexicon.prl > ${LEXICON_TRAIN};
cat ${TRAIN} | ${DIR_UTIL}/make-restricted-lexicon.prl > ${LEXICON_ALL};

echo "Create untagged";
cat ${TRAIN_2} | $DIR_UTIL/tagged-to-untagged.prl > ${TRAIN_UNTAGGED_2};

echo "BRILL - Create dummy tagger";
cd ${DIR_BIN};
./tagger ${LEXICON_TRAIN} ${TRAIN_UNTAGGED_2} ${BIGRAMS_TRAIN} ${RULES_TAGS} /dev/null -w ${LIST_BIG} -i ${TAGGED_DUMMY} > /dev/null

echo "BRILL - Learn contextual rules";
if [ ! -f ${RULES_CONTEXTUAL} ]; then
	cd ${DIR_BIN};
	./contextual-rule-learn ${TRAIN_2} ${TAGGED_DUMMY} ${RULES_CONTEXTUAL} ${LEXICON_TRAIN};
fi;



echo "BRILL - Stripping tags from TRAIN"
cd ${DIR_DATA};
cat ${TEST} | perl ${DIR_UTIL}/tagged-to-untagged.prl > ${TEST_UNTAGGED};

${DIR_UTIL}/incorporate-new-bigrams.prl ${RULES_TAGS} ${BIGRAMS_TRAIN} ${TEST_UNTAGGED} > ${BIGRAMS_ALL}

${DIR_UTIL}/incorporate-new-words.prl ${LEXICON_ALL} ${TEST_UNTAGGED} > ${WORDS}

echo "BRILL - Tag test";
cd ${DIR_BIN};
./tagger ${LEXICON_ALL} ${TEST_UNTAGGED} ${BIGRAMS_ALL} ${RULES_TAGS} ${RULES_CONTEXTUAL} -w ${WORDS} > ${TAGGED_RESULT}

cd ${DIR_ACT};

echo "BRILL - Evaluate";
${DIR_ACT}/eval.pl ${TEST} ${TAGGED_RESULT} | tee ${RESULT};

