#!/usr/bin/perl
use strict;
use warnings;
use POSIX;

use LM::Utils;


if ( ! defined($ARGV[0]) ) { 
	print "Usage: ./eval.pl EXPECTED_CORPUS ACTUAL_CORPUS [limit]";
}

my $fileExp = $ARGV[0];
my $fileAct = $ARGV[1];
my $limit   = $ARGV[2];
if ( ! defined($limit) ) { 
	$limit = POSIX::INT_MAX;
}

open FH_EXP, $fileExp;
open FH_ACT, $fileAct;

my $wordTotal = 0;
my $wordCorrect = 0;
my $sentenceTotal = 0;
my $sentenceCorrect = 0;
my %bugs;

while (<FH_EXP>) { 
	my $lineExp = LM::Utils::trim($_);
	my $lineAct = <FH_ACT>;
	$lineAct = LM::Utils::trim($lineAct);
	
	my @wordsExp = split(/ /, $lineExp);
	my @wordsAct = split(/ /, $lineAct);
	
	my $isSentceCorrect = 1;
	for my $i (0..$#wordsExp) { 
		my @pExp = split(/\//, $wordsExp[$i]);
		my @pAct = split(/\//, $wordsAct[$i]);
		
		if ( $pExp[1] eq $pAct[1] ) { 
			$wordCorrect++;
		} else { 
			$isSentceCorrect = 0;
			$bugs{$pExp[1] . "\t" . $pAct[1]}++;	
		}
		
		$wordTotal++;
	}
	
	$sentenceCorrect += $isSentceCorrect;
	$sentenceTotal++;	
}

print "W:\t" . $wordCorrect . "\t" . $wordTotal . "\t"  . (100.0 * $wordCorrect / $wordTotal) . "\n";
print "S:\t" . $sentenceCorrect . "\t" . $sentenceTotal . "\t"  . (100.0 * $sentenceCorrect / $sentenceTotal) . "\n";
print "\n";
print "Exp\tWas\tCount\n";
my $i = 0;
for my $k (sort { $bugs{$b} <=> $bugs{$a}; } keys %bugs) {
	if ( $i >= $limit ) { 
		last;
	} 
	print $k, "\t", $bugs{$k}, "\n";
	
	$i++;
}
  
