#!/usr/bin/perl

use utf8;
use strict;
use warnings;


use LM::Smooth;
use LM::Utils;
use LM::Viterbi;

sub readFile {
	my($fileName) = shift; 
	my(@text); 
	open(FILE, "<",$fileName);
	while(<FILE>){
		my @a = split(/\s+/, $_);
		push @text, ["###","###"];
		foreach(@a){
			my @p = split(/\//, $_);
			push @text, \@p;
		}		
	}
	close FILE;
	return \@text;
}

sub help()
{
	print STDERR "Usage: ./hmm-tagger.pl TRAIN HELDOUT TEST [lambdasLex] [lambdasTri]\n";
	exit;
}

if ( ! defined($ARGV[2])) { 
	help();
	exit;
}
 
my $fTrain = $ARGV[0];
if ( ! -f $fTrain || ! -r $fTrain ) { 
	print STDERR "Train file $fTrain is not readable.\n";
	exit;
}

my $fHeldout = $ARGV[1];
if ( ! -f $fHeldout || ! -r $fHeldout ) { 
	print STDERR "Held out file $fHeldout is not readable.\n";
	exit;
}


my $fTest = $ARGV[2];
if ( ! -f $fTest || ! -r $fTest ) { 
	print STDERR "Test file $fTest is not readable.\n";
	exit;
}

my $fLambdaLex;
if ( $ARGV[3] ) { 
	$fLambdaLex = $ARGV[3];
} else { 
	$fLambdaLex = $fTrain . '.lambda.lex';
}

my $fLambdaTri;
if ( $ARGV[4] ) { 
	$fLambdaTri = $ARGV[4];
} else { 
	$fLambdaTri = $fTrain . '.lambda.tri';
}



my $rTrain = readFile($fTrain);
my $rHeldout = readFile($fHeldout);

# trigramovy model
my $smoothTri = new LM::Smooth(); 
$smoothTri->prob($rTrain); 
if(-r $fLambdaTri){
	$smoothTri->loadLambdas($fLambdaTri);
}
else {
	$smoothTri->smooth($rHeldout, 0.0001);
	$smoothTri->saveLambdas($fLambdaTri);
}
LM::Utils::dbgLog("Compressing");
$smoothTri->shrink();


# lexikalni model	
my $smoothLex = new LM::Smooth(1); 
$smoothLex->prob($rTrain); 

my $half = int($#{$rHeldout}/2);
my @heldUnk = @{$rHeldout}[0..$half-1];
my @heldRest = @{$rHeldout}[$half..$#{$rHeldout}];
$smoothLex->smoothUnknown(\@heldUnk);

if(-r $fLambdaLex){
	$smoothLex->loadLambdas($fLambdaLex);
}
else {
	$smoothLex->smooth(\@heldRest, 0.0001);
	$smoothLex->saveLambdas($fLambdaLex);
}

LM::Utils::dbgLog("Compressing");
$smoothLex->shrink();

my $viterbi = LM::Viterbi->new($smoothTri, $smoothLex);

LM::Utils::dbgLog("Running Viterbi");
$viterbi->run($fTest, 15); 
