#!/usr/bin/perl
use utf8;
use strict;
use warnings;

use LM::Utils;

my $TEST_SIZE = 40000;
my $HELDOUT_SIZE = 20000;
my $TEST_SUFFIX = "test";
my $HELDOUT_SUFFIX = "heldout";
my $TRAIN_SUFFIX = "train";
my $file = $ARGV[0];

open FH, $file;
my @sentenceLengths = ();
my @sentences = ();
my @sentence = ();
my $ln = 0; 
while (<FH>) { 
	my @seq = split(/\//, LM::Utils::trim($_));
	if ( $seq[0] eq "###") { 
		if ( @sentence ) { 
			my @tmp = @sentence;
			push(@sentences, \@tmp);
			push(@sentenceLengths, ($#tmp) + 1);
			@sentence = ();			
		}
	} else { 
		push(@sentence, \@seq);
		$ln++;	
	}
}
close(FH);

print STDERR "Number of sentences: ", $#sentences, "\n";

my ($from, $to);
($from, $to) = createFile("01." . $TEST_SUFFIX , $ln - 1.01 * $TEST_SIZE, $TEST_SIZE);

($from, $to) = createFile("01." . $HELDOUT_SUFFIX, $from - $HELDOUT_SIZE, $HELDOUT_SIZE);
($from, $to) = createFile("01." . $TRAIN_SUFFIX, 0, $from);


($from, $to) = createFile("02." . $TEST_SUFFIX, 0, $TEST_SIZE);
($from, $to) = createFile("02." . $HELDOUT_SUFFIX, $to, $HELDOUT_SIZE);
($from, $to) = createFile("02." . $TRAIN_SUFFIX, $to, $ln - $HELDOUT_SIZE - $TEST_SIZE);

createSet("03", $TEST_SIZE, $HELDOUT_SIZE);
createSet("04", $TEST_SIZE, $HELDOUT_SIZE);
createSet("05", $TEST_SIZE, $HELDOUT_SIZE);


sub createSet
{
	my ($suffix, $testSize, $heldoutSize) = @_;
	
	my @test = ();
	my @heldout = ();
	my @train = ();
	my ($actTest, $actHeldout, $actTrain) = (0, 0, 0);
	my $sentecePos = 0;
	for my $l ( @sentenceLengths ) {
		my $r = rand();
		if ( $r < 0.333 && $actTest < $testSize ) { 
			push(@test, $sentences[$sentecePos]);
			$actTest += $l;
		} elsif ( $r < 0.666 && $actHeldout < $heldoutSize ) { 
			push(@heldout, $sentences[$sentecePos]);
			$actHeldout += $l;
		} else { 
			push(@train, $sentences[$sentecePos]);
		}		
		
		$sentecePos++;
	}
	save($file . "." . $suffix . "." . $TEST_SUFFIX , \@test);
	save($file . "." . $suffix . "." . $TRAIN_SUFFIX, \@train);
	save($file . "." . $suffix . "." . $HELDOUT_SUFFIX, \@heldout);	
	
	
}
sub createFile
{
	my ($suffix, $from, $size) = @_;
	print STDERR "$suffix; From: $from; Size: $size\n";
	my $actLen = 0;
	my ($rFrom, $rTo) = (undef, undef);
	my $wordPos = 0;
	my $sentencePos = 0;
	my @res = ();
	for my $l ( @sentenceLengths ) {
		if ( $actLen > $size ) { 
			last;
		} elsif ( $wordPos >= $from ) { 
			if ( ! defined($rFrom) ) { 
				$rFrom = $wordPos;
			}
			$rTo = $wordPos;
			push(@res, $sentences[$sentencePos]);
			$actLen += $l;			
		}
		
		$wordPos = $wordPos + $l;
		$sentencePos++;
	} 
	print STDERR "Read from $rFrom to $rTo\n";
	save($file . "." . $suffix, \@res);
	
	return ($rFrom, $rTo + 1);
	
}

sub save
{
	my ($file, $res) = @_;
	open FH,  ">", $file;
	for my $r (@{$res}) { 
		my @p = ();
		for my $w (@{$r}) {
			push(@p, $w->[0] . "/" . $w->[1]); 
		}
		print FH join(" ", @p);
		print FH "\n";		
	}
	close(FH);
	
	
}