#!/usr/bin/perl
use strict;
use warnings;
use utf8;
use Data::Dumper;

binmode STDIN, ':utf8';
binmode STDOUT, ':utf8';
binmode STDERR, ':utf8';

use LM::NGramLangModel;
use LM::Utils;

my $file = $ARGV[0];
my $window = $ARGV[1];
if ( ! defined($window) ) { 
	$window = 50;
}
my $n = 2;

my $model = LM::NGramLangModel->new($n);


print STDERR "Processing input\n";

my $fh;
open($fh, "<utf8:", $file);
my @lines = ();
while (my $line = <$fh>) { 
	$line = LM::Utils::trim($line);
	push(@lines, $line);
}
close($fh);

my $max = (@lines) - 1;

my @nGram;
my %words = ();
my $wordCount = 0;
for my $i (0..$max ) { 
	my $from = LM::Utils::max($i - $window, 0);
	my $to = LM::Utils::min($i + $window, $max);
	$words{$lines[$i]}++;	
	for (my $j = $from; $j <= $to; $j++) {
		if ( $j > $i - 2 && $j < $i + 2 ) { 
			next;
		}
		
		if ( $j < $i ) {
			$nGram[0] = $lines[$j];
			$nGram[1] = $lines[$i];
		} else { 
			$nGram[0] = $lines[$i];
			$nGram[1] = $lines[$j];
		}
		
		$model->addNGram(\@nGram);		
	}
}

#print STDERR "Filtering\n";	
#$model->filter(\&filter);

print STDERR "Computing counts\n";
$model->computeCounts();

print STDERR "Computing probs\n";
$model->computeProbs(1);

print STDERR "Printing ngrams\n";
my @nGrams = $model->getAllNGrams();
for my $nGram (@nGrams) { 
	if ( $words{$nGram->[0]} >= 10 && $words{$nGram->[1]} >= 10 ) {
		print join("\t", @{$nGram}), "\t", $model->getPointwiseMutualInformation($nGram), "\n";
	} 
}


sub filter { 
	my $count = shift;
	return $count < 10;
}