#!/usr/bin/perl
use strict;
use warnings;
no warnings 'uninitialized';

use Data::Dumper;
binmode STDIN, ':utf8';
binmode STDOUT, ':utf8';
binmode STDERR, ':utf8';

sub help
{
	print STDERR "Usage: ./task2-optimal.pl tags/words THRESHOLD FILENAME [merges/classes] [class_count]\n";
}

use LM::Utils;

my $inputType = $ARGV[0];
my $threshold = $ARGV[1];
my $fileName = $ARGV[2];
my $outputType = $ARGV[3];
my $thresholdClassCount = $ARGV[4];

if ( ! $outputType ) { 
	$outputType = "merges";
}

if ( ! $fileName ) { 
	help();
	exit;
}

my %counts;

if ( ! -f $fileName || ! -r $fileName ) { 
	print STDERR "Soubor $fileName neni mozne cist.\n";
	help();
	exit;
}

my $readPos = 0;
if ( $inputType eq "tags" ) { 
	$readPos = 1;
} elsif ( $inputType ne "words" ) { 
	help();
	exit;	
}

if ( $outputType ne "merges" && $outputType ne "classes") { 
	help();
	exit;	
}

if ( $outputType eq "classes" && ! $thresholdClassCount ) { 
	$thresholdClassCount = 15;
}

# data structures
my $mutualInformation = 0;
my $bigramCount = 0;	
my $wordCount = 0;
my @words = ();			#
my %wordMap = ();		#
my %bigrams = ();		# c
my %bigramsOld = ();
my %unigramsL = ();		# cL
my %unigramsLOld = ();
my %unigramsR = ();		# cR
my %unigramsROld = ();
my %subterms = ();		# q
my %subtermsOld = ();
my %subtractions = ();	# s
my %subtractionsOld = ();
my %loses = ();			# L
my %losesOld = ();			# L
my $N = 0;
my @history = ();
my %classes = ();

my $minLoss = 100;
my $minA = -1;
my $minWA = "";
my $minWAOld = "";
my $minB = -1;
my $minWB = "";
my $minWBOld = "";
my $cLAB = 0;
my $cRAB = 0;


# slide 187

# 1. read data + init bigrams + words
LM::Utils::dbgLog("Phase1 - BEGIN");
open(FH, "<utf8:", $fileName);
my @parts;
my $last = "__BEGIN__";
my $act;

while (<FH>) { 
	@parts = split(/\//, LM::Utils::trim($_));
	
	$act = $parts[$readPos];
	$bigrams{$last}{$act}++;
	
	$last = $act;	
	$bigramCount++;
}
LM::Utils::dbgLog("Bigram count: $bigramCount");

LM::Utils::dbgLog("Phase1 - END");


# 2. init unigrams
LM::Utils::dbgLog("Phase2 - BEGIN");
for my $l ( keys %bigrams ) { 
	for my $r ( keys %{$bigrams{$l}} ) {
		$unigramsL{$l} += $bigrams{$l}{$r};
		$unigramsR{$r} += $bigrams{$l}{$r};
	}
}
$N = scalar keys %unigramsR;
LM::Utils::dbgLog("Phase2 - END");

# select just frequent words
LM::Utils::dbgLog("Phase2 b - frequent words - BEGIN");
for my $w ( sort keys %unigramsR) {
	if ( $unigramsR{$w} >= $threshold ) {
#		print STDERR $w, "\n"; 
		push(@words, $w);
		$wordMap{$w} = $wordCount;
		$wordCount++;
		$classes{$w} = "(" . $w . ")";
	}	
}
LM::Utils::dbgLog("Frequent word count: $wordCount");
LM::Utils::dbgLog("Phase2 b - frequent words - END");

# 3. init subterms
LM::Utils::dbgLog("Mutual Information - INIT - BEGIN");
$mutualInformation = 0;
for my $l ( keys %bigrams ) { 
	for my $r ( keys %{$bigrams{$l}} ) {
		$subterms{$l}{$r} = ( $bigrams{$l}{$r} / $bigramCount ) * 
			LM::Utils::log2( ( $bigramCount * $bigrams{$l}{$r}) /
				 ( $unigramsL{$l} * $unigramsR{$r})
		
			);
		$mutualInformation += $subterms{$l}{$r};
	}
}	
LM::Utils::dbgLog("Mutual Information: $mutualInformation");
LM::Utils::dbgLog("Mutual Information - INIT - END");

# 4. init subtraction
LM::Utils::dbgLog("Subtractions - BEGIN");
for my $a (@words) {
	my $sumL = 0;
	for my $l (keys %unigramsL) {
		$sumL += $subterms{$l}{$a};
	}
	
	my $sumR = 0;
	for my $r (keys %unigramsR) {
		$sumR += $subterms{$a}{$r};
	}
	
	$subtractions{$a} = $sumL + $sumR - $subterms{$a}{$a};	
#	print STDERR "$a = ", $subtractions{$a}, "\n"; 
}
LM::Utils::dbgLog("Subtractions - END");

# 5. init table of loses
LM::Utils::dbgLog("Losses - BEGIN");
$minLoss = 100;
$minWA = "";
$minWB = "";
$cLAB = 0;
$cRAB = 0;

my ($sub, $add, $loss); 
my $i = 0;
my $j = 0;
for my $wA (@words) {
	my $j = 0;
		
	for my $wB (@words) {
		$j++;
		if ( $j > $i ) { 
			next;
		}
		
		$sub = $subtractions{$wA} + $subtractions{$wB} - 
			   $subterms{$wA}{$wB} - $subterms{$wB}{$wA};
			   
			   
		my $sumL = 0;
		$cRAB = $unigramsR{$wA} + $unigramsR{$wB};
		for my $l (keys %unigramsL) {
			if ( $l eq $wA || $l eq $wB ) { 
				next;
			}
			$sumL += getSubterms(
				$bigramCount, 
				($bigrams{$l}{$wA} + $bigrams{$l}{$wB}),
				$unigramsL{$l}, 
				$cRAB 
				);
		}
		
		my $sumR = 0;
		$cLAB = $unigramsL{$wA} + $unigramsL{$wB};
		for my $r (keys %unigramsR) {
			if ( $r eq $wA || $r eq $wB ) { 
				next;
			}
			$sumR += getSubterms(
				$bigramCount, 
				($bigrams{$wA}{$r} + $bigrams{$wB}{$r}),
				$unigramsR{$r}, 
				$cLAB 
				);
		}
		
		$add = $sumL + $sumR + getSubterms(
				$bigramCount, 
				($bigrams{$wA}{$wA} + $bigrams{$wA}{$wB} + 
				 $bigrams{$wB}{$wA} + $bigrams{$wB}{$wB}),
				$cLAB, 
				$cRAB 
				);	   		
		
		$loss = $sub - $add;
		
		
		
		$loses{$wA}{$wB} = $loss;
						 
		if ( $loss < $minLoss ) { 
			$minLoss = $loss;
			$minWA = $wA;
			$minWB = $wB;			
		}
					
	}
	$i++;
}
LM::Utils::dbgLog("Losses - END");	



while ( $wordCount > 1) {	
	
	# save to history
	LM::Utils::dbgLog("Merging - BEGIN");
	{
		my @rec = ($minWA, $minWB, $minLoss);
		push(@history, \@rec);
	}

	# merge unigrams
	%unigramsLOld = %unigramsL;
	$unigramsL{$minWA} = $unigramsL{$minWA} + $unigramsR{$minWB};
	delete $unigramsL{$minWB};
	
	%unigramsROld = %unigramsR;
	$unigramsR{$minWA} = $unigramsR{$minWA} + $unigramsR{$minWB};
	delete $unigramsR{$minWB};
	
	# merge bigrams
	%bigramsOld = %bigrams;
	for my $l (keys %bigrams) { 
		if ( $l eq $minWB ) { 
			for my $r (keys %{$bigrams{$l}}) { 
				if ( $r eq $minWB ) { 
					$bigrams{$minWA}{$minWA} += $bigrams{$l}{$minWB}; 
				} else { 
					$bigrams{$minWA}{$r} += $bigrams{$l}{$r};
				}
			}
		} else {
			if ( exists($bigrams{$l}{$minWB}) ) { 
				$bigrams{$l}{$minWA} += $bigrams{$l}{$minWB}; 
			}
			delete $bigrams{$l}{$minWB};
		}
	}
	delete $bigrams{$minWB};
	$bigramCount--;	
	
	# merge word map
	delete $wordMap{$minWB};
	@words = ();
	$wordCount = 0;
	for my $w (sort keys %wordMap) { 
		push(@words, $w);
		$wordMap{$w} = $wordCount;
		$wordCount++;
	}

	# merge classes
	$classes{$minWA} = "(" . $classes{$minWA} . " + " . $classes{$minWB} . ")";
	undef $classes{$minWB};
	
	LM::Utils::dbgLog("MERGE: $minWA\t$minWB\t$minLoss");
	LM::Utils::dbgLog("CLASSES: $classes{$minWA}");
	if ( $outputType eq "merges" ) {
		print "$wordCount\t$minWA\t$minWB\t$mutualInformation\t$minLoss\t$classes{$minWA}\n";
	} elsif ( $outputType eq "classes" && $wordCount == $thresholdClassCount ) { 
		for my $k (keys %classes) {
			if ( $classes{$k} ) { 
				print $classes{$k}, "\n";
			}
		}
		exit;
	}
	
	LM::Utils::dbgLog("Merging - END");
	
	# 3. update subterms
	LM::Utils::dbgLog("Mutual Information - UPDATE - BEGIN");
	
	%subtermsOld = %subterms;
	$mutualInformation = 0;
	for my $l ( keys %bigrams ) { 
		for my $r ( keys %{$bigrams{$l}} ) {
			$subterms{$l}{$r} = ( $bigrams{$l}{$r} / $bigramCount ) * 
				LM::Utils::log2( ( $bigramCount * $bigrams{$l}{$r}) /
					 ( $unigramsL{$l} * $unigramsR{$r})
			
				);
			$mutualInformation += $subterms{$l}{$r};
		}

	}	
	LM::Utils::dbgLog("Mutual Information: $mutualInformation");
	LM::Utils::dbgLog("Mutual Information - UPDATE - END");	
	
	# 4. init subtraction
	LM::Utils::dbgLog("Subtractions - UPDATE - BEGIN");
	my $a;
	%subtractionsOld = %subtractions;
=iem	
	for my $i (@words) {
		$subtractions{$i} = $subtractionsOld{$i} 
			- $subtermsOld{$i}{$minWA} - $subtermsOld{$minWA}{$i}
			- $subtermsOld{$i}{$minWB} - $subtermsOld{$minWB}{$i}
			+ $subterms{$i}{$minWA} + $subterms{$minWA}{$i}
			;	
		print STDERR "$i = ", $subtractions{$i}, "($subterms{$i}{$minWA}, $subterms{$minWA}{$i})\n"; 
	}
=cut	

	for my $a (@words) {
		my $sumL = 0;
		for my $l (keys %unigramsL) {
			$sumL += $subterms{$l}{$a};
		}
		
		my $sumR = 0;
		for my $r (keys %unigramsR) {
			$sumR += $subterms{$a}{$r};
		}
		
		$subtractions{$a} = $sumL + $sumR - $subterms{$a}{$a};	
#		print STDERR "$a = ", $subtractions{$a}, "\n"; 
	}

	LM::Utils::dbgLog("Subtractions - UPDATE - END"); 
	
		
	LM::Utils::dbgLog("Losses - UPDATE - BEGIN");
	$minLoss = 100;
	
	$minWAOld = $minWA;
	$minWBOld = $minWB;
	$minWA = "";
	$minWB = "";	
	$cLAB = 0;
	$cRAB = 0;
	my ($sub, $add, $loss); 
	my $i = 0;
	my $j = 0;
	for my $wA (@words) {
		my $j = 0;
		if ( $wA eq $minWAOld ) { 
			$i++;
			next;
		}
			
		for my $wB (@words) {
			$j++;
			if ( $j > $i ) { 
				next;
			}			
			
			if ( $wB eq $minWAOld ) { 
				next;
			}
	
			
			$loss = $loses{$wA}{$wB} - 
				$subtractionsOld{$wA} + $subtractions{$wA} -
				$subtractionsOld{$wB} + $subtractions{$wB} + 
				getSubterms(
					$bigramCount, 
					($bigrams{$wA}{$minWAOld} + $bigrams{$wB}{$minWAOld}),
					$unigramsL{$wA} + $unigramsL{$wB}, 
					$unigramsR{$minWAOld} 
					) + 
				getSubterms(
					$bigramCount, 
					($bigrams{$minWAOld}{$wA} + $bigrams{$minWAOld}{$wB}),
					$unigramsL{$minWAOld}, 
					$unigramsR{$wA} + $unigramsR{$wB}				 
					) + 
				getSubterms(
					$bigramCount, 
					($bigrams{$wA}{$minWBOld} + $bigrams{$wB}{$minWBOld}),
					$unigramsL{$wA} + $unigramsL{$wB}, 
					$unigramsR{$minWBOld} 
					) + 
				getSubterms(
					$bigramCount, 
					($bigrams{$minWBOld}{$wA} + $bigrams{$minWBOld}{$wB}),
					$unigramsL{$minWBOld}, 
					$unigramsR{$wA} + $unigramsR{$wB}				 
					) - 
				getSubterms(
					$bigramCount, 
					($bigrams{$wA}{$minWAOld} + $bigrams{$wB}{$minWAOld}),
					$unigramsL{$wA} + $unigramsL{$wB}, 
					$unigramsR{$minWAOld} 
					) - 
				getSubterms(
					$bigramCount, 
					($bigrams{$minWAOld}{$wA} + $bigrams{$minWAOld}{$wB}),
					$unigramsL{$minWAOld}, 
					$unigramsR{$wA} + $unigramsR{$wB}				 
					);
			
			
			
			$loses{$wA}{$wB} = $loss;
							 
			if ( $loss < $minLoss && $loss > 0 ) { 
				$minLoss = $loss;
				$minWA = $wA;
				$minWB = $wB;		
#				LM::Utils::dbgLog("MIN: $wA + $wB\t$minLoss");	
			}
								
		}
		$i++;
	}
	

	my $wA = $minWAOld;
	for my $wB (@words) {
		if ( $wB eq $wA ) { 
			next;
		}
		
		$sub = $subtractions{$wA} + $subtractions{$wB} - 
			   $subterms{$wA}{$wB} - $subterms{$wB}{$wA};
			   
			   
		my $sumL = 0;
		$cRAB = $unigramsR{$wA} + $unigramsR{$wB};
		for my $l (keys %unigramsL) {
			if ( $l eq $wA || $l eq $wB ) { 
				next;
			}
			$sumL += getSubterms(
				$bigramCount, 
				($bigrams{$l}{$wA} + $bigrams{$l}{$wB}),
				$unigramsL{$l}, 
				$cRAB 
				);
		}
		
		my $sumR = 0;
		$cLAB = $unigramsL{$wA} + $unigramsL{$wB};
		for my $r (keys %unigramsR) {
			if ( $r eq $wA || $r eq $wB ) { 
				next;
			}
			$sumR += getSubterms(
				$bigramCount, 
				($bigrams{$wA}{$r} + $bigrams{$wB}{$r}),
				$unigramsR{$r}, 
				$cLAB 
				);
		}
		
		$add = $sumL + $sumR + getSubterms(
				$bigramCount, 
				($bigrams{$wA}{$wA} + $bigrams{$wA}{$wB} + 
				 $bigrams{$wB}{$wA} + $bigrams{$wB}{$wB}),
				$cLAB, 
				$cRAB 
				);	   		
		
		$loss = $sub - $add;
		
		
		
		$loses{$wA}{$wB} = $loss;
						 
		if ( $loss < $minLoss && $loss > 0) { 
			$minLoss = $loss;
			$minWA = $wA;
			$minWB = $wB;	
#			LM::Utils::dbgLog("New - MIN: $wA + $wB\t$minLoss");	
		}
					
	}
	$i++;	
	
	LM::Utils::dbgLog("Losses - UPDATE - END");

	LM::Utils::dbgLog("Iterations to go $wordCount");	
}

	
#print Data::Dumper->Dump([\@history]);

sub getSubterms
{
	my ($N, $a, $b, $c) = @_;
	if ( $a == 0 || $b == 0 || $c == 0 ) { 
		return 0;
	}
	return ($a / $N) * LM::Utils::log2(($N * $a) / ($b * $c));
	
}

