#!/usr/bin/perl
use strict;
use warnings;
no warnings 'uninitialized';

use Data::Dumper;
binmode STDIN, ':utf8';
binmode STDOUT, ':utf8';
binmode STDERR, ':utf8';


use lib "/home/martin/school/libs/perl";
use LM::Utils;

my $inputType = $ARGV[0];
my $threshold = $ARGV[1];
my $fileName = $ARGV[2];

my %counts;

my $readPos = 0;
if ( $inputType eq "tags" ) { 
	$readPos = 1;
} elsif ( $inputType ne "words" ) { 
	help();
	exit;	
}

# data structures
my $mutualInformation = 0;
my $bigramCount = 0;	
my $wordCount = 0;
my @words = ();			#
my %wordMap = ();		#
my %bigrams = ();		# c
my %unigramsL = ();		# cL
my %unigramsR = ();		# cR
my %subterms = ();		# q
my @subtractions = ();	# s
my @loses = ();			# L
my $N = 0;
my @history = ();
my %classes = ();

my $minLoss = 100;
my $minA = -1;
my $minWA = "";
my $minB = -1;
my $minWB = "";
my $cLAB = 0;
my $cRAB = 0;


# slide 187

# 1. read data + init bigrams + words
LM::Utils::dbgLog("Phase1 - BEGIN");
open(FH, "<utf8:", $fileName);
my @parts;
my $last = "__BEGIN__";
my $act;

while (<FH>) { 
	@parts = split(/\//, LM::Utils::trim($_));
	
	$act = $parts[$readPos];
	$bigrams{$last}{$act}++;
	
	$last = $act;	
	$bigramCount++;
}
LM::Utils::dbgLog("Bigram count: $bigramCount");

LM::Utils::dbgLog("Phase1 - END");


# 2. init unigrams
LM::Utils::dbgLog("Phase2 - BEGIN");
for my $l ( keys %bigrams ) { 
	for my $r ( keys %{$bigrams{$l}} ) {
		$unigramsL{$l} += $bigrams{$l}{$r};
		$unigramsR{$r} += $bigrams{$l}{$r};
	}
}
$N = scalar keys %unigramsR;
LM::Utils::dbgLog("Phase2 - END");

# select just frequent words
LM::Utils::dbgLog("Phase2 b - frequent words - BEGIN");
for my $w ( keys %unigramsR) {
	if ( $unigramsR{$w} >= $threshold ) { 
		push(@words, $w);
		$wordMap{$w} = $wordCount;
		$wordCount++;
		$classes{$w} = "(" . $w . ")";
	}	
}
LM::Utils::dbgLog("Frequent word count: $wordCount");
LM::Utils::dbgLog("Phase2 b - frequent words - END");





do {	
	LM::Utils::dbgLog("Iterations to go $wordCount");
	# 3. init subterms
	LM::Utils::dbgLog("Mutual Information - BEGIN");
	$mutualInformation = 0;
	for my $l ( keys %bigrams ) { 
		for my $r ( keys %{$bigrams{$l}} ) {
			$subterms{$l}{$r} = ( $bigrams{$l}{$r} / $bigramCount ) * 
				LM::Utils::log2( ( $bigramCount * $bigrams{$l}{$r}) /
					 ( $unigramsL{$l} * $unigramsR{$r})
			
				);
			$mutualInformation += $subterms{$l}{$r};
		}
	}	
	LM::Utils::dbgLog("Mutual Information: $mutualInformation");
	LM::Utils::dbgLog("Mutual Information - END");
	
	# 4. init subtraction
	LM::Utils::dbgLog("Subtractions - BEGIN");
	my $a;
	for my $i (0..($wordCount - 1)) {
		$a = $words[$i]; 
		my $sumL = 0;
		for my $l (keys %unigramsL) {
			$sumL += $subterms{$l}{$a};
		}
		
		my $sumR = 0;
		for my $r (keys %unigramsR) {
			$sumR += $subterms{$a}{$r};
		}
		
		$subtractions[$i] = $sumL + $sumR - $subterms{$a}{$a};	
#		print STDERR "$a = ", $subtractions[$i], "($subterms{$a}{$minWA}, $subterms{$minWA}{$a})\n"; 
	}
	LM::Utils::dbgLog("Subtractions - END");
	
	# 5. init table of loses
	LM::Utils::dbgLog("Losses - BEGIN");
	$minLoss = 100;
	$minA = -1;
	$minWA = "";
	$minB = -1;
	$minWB = "";
	$cLAB = 0;
	$cRAB = 0;
	
	my ($sub, $add, $loss); 
	for my $a (0..($wordCount-1)) {
		my $wA = $words[$a];
			
		for my $b (($a+1) .. ($wordCount-1)) { 
			my $wB = $words[$b];
			
			$sub = $subtractions[$a] + $subtractions[$b] - 
				   $subterms{$wA}{$wB} - $subterms{$wB}{$wA};
				   
				   
			my $sumL = 0;
			$cRAB = $unigramsR{$wA} + $unigramsR{$wB};
			for my $l (keys %unigramsL) {
				if ( $l eq $wA || $l eq $wB ) { 
					next;
				}
				$sumL += getSubterms(
					$bigramCount, 
					($bigrams{$l}{$wA} + $bigrams{$l}{$wB}),
					$unigramsL{$l}, 
					$cRAB 
					);
			}
			
			my $sumR = 0;
			$cLAB = $unigramsL{$wA} + $unigramsL{$wB};
			for my $r (keys %unigramsR) {
				if ( $r eq $wA || $r eq $wB ) { 
					next;
				}
				$sumR += getSubterms(
					$bigramCount, 
					($bigrams{$wA}{$r} + $bigrams{$wB}{$r}),
					$unigramsR{$r}, 
					$cLAB 
					);
			}
			
			$add = $sumL + $sumR + getSubterms(
					$bigramCount, 
					($bigrams{$wA}{$wA} + $bigrams{$wA}{$wB} + 
					 $bigrams{$wB}{$wA} + $bigrams{$wB}{$wB}),
					$cLAB, 
					$cRAB 
					);	   		
			
			$loss = $sub - $add;
			
			
			
			$loses[$a][$b] = $loss;
							 
			if ( $loss < $minLoss ) { 
				$minLoss = $loss;
				$minA = $a;
				$minB = $b;
				$minWA = $wA;
				$minWB = $wB;			
			}
						
		}
	}
	LM::Utils::dbgLog("Losses - END");	
	
	
	# save to history
	LM::Utils::dbgLog("Merging- BEGIN");
	{
		my @rec = ($minWA, $minWB, $minLoss);
		push(@history, \@rec);
	}

	# merge unigrams
	$unigramsL{$minWA} = $unigramsL{$minWA} + $unigramsR{$minWB};
	delete $unigramsL{$minWB};
	$unigramsR{$minWA} = $unigramsR{$minWA} + $unigramsR{$minWB};
	delete $unigramsR{$minWB};
	
	# merge bigrams
	for my $l (keys %bigrams) { 
		if ( $l eq $minWB ) { 
			for my $r (keys %{$bigrams{$l}}) { 
				if ( $r eq $minWB ) { 
					$bigrams{$minWA}{$minWA} += $bigrams{$l}{$minWB}; 
				} else { 
					$bigrams{$minWA}{$r} += $bigrams{$l}{$r};
				}
			}
		} else { 
			for my $r (keys %{$bigrams{$l}}) { 
				if ( $r eq $minWB ) { 
					$bigrams{$l}{$minWA} += $bigrams{$l}{$minWB}; 
				}
			}
			delete $bigrams{$l}{$minWB};
		}
	}
	delete $bigrams{$minWB};
	$bigramCount--;	
	
	# merge word map
	delete $wordMap{$minWB};
	@words = ();
	$wordCount = 0;
	for my $w (keys %wordMap) { 
		push(@words, $w);
		$wordMap{$w} = $wordCount;
		$wordCount++;
	}
	
	# merge classes
	$classes{$minWA} = "(" . $classes{$minWA} . " + " . $classes{$minWB} . ")";
	
	LM::Utils::dbgLog("MERGE: $minWA\t$minWB\t$minLoss");
	LM::Utils::dbgLog("CLASSES: $classes{$minWA}");
	
	print "$wordCount\t$minWA\t$minWB\t$mutualInformation\t$minLoss\t$classes{$minWA}\n";
	
	LM::Utils::dbgLog("Merging- END");
	
	
} while( $wordCount > 1);

	
#print Data::Dumper->Dump([\@history]);

sub getSubterms
{
	my ($N, $a, $b, $c) = @_;
	if ( $a == 0 || $b == 0 || $c == 0 ) { 
		return 0;
	}
	return ($a / $N) * LM::Utils::log2(($N * $a) / ($b * $c));
	
}

